#!/usr/bin/python -tt

import posix
import sys
import os
import re
import syslog
import subprocess

default_disk_root = '/srv'
default_mount_params = 'rw,nosuid,nodev,noatime'
default_allow_mount = 1
config_file = '/etc/default/gandi'
datadisk_root = ''


def load_mount_params(config_file):
    """ Load mount parameter from config file """
    return load_config_file(config_file,
                r'^CONFIG_MOUNT_PARAMS.?=([^ #]*)$',
                default_mount_params,
                'CONFIG_MOUNT_PARAMS')


def load_disk_root(config_file):
    """ Load disk root parameter from config file """
    return load_config_file(config_file,
                r'^CONFIG_DISK_ROOT.?=([^ #]*)$',
                default_disk_root,
                'CONFIG_DISK_ROOT')


def load_allow_mount(config_file):
    """ Load flag to enable or disable mount from config file """
    return load_config_file(config_file,
            r'^CONFIG_ALLOW_MOUNT.?=([^ #]*)$',
            default_allow_mount,
            'CONFIG_ALLOW_MOUNT')


def load_config_file(config_file, pattern, default_value, label):
    """ Load specific options from config_file"""
    params = default_value
    patternc = re.compile(pattern)

    if not os.path.exists(config_file):
        config_file='/etc/sysconfig/gandi'

    try:
        for line in open(config_file, 'r'):
            match = patternc.search(line.rstrip(os.linesep))
            if match is not None:
                defined_value = match.group(1).strip('"')
                if defined_value != "":
                    params = defined_value
                else:
                    syslog.syslog("%s is empty. Using defaults : %s." % \
                        (label, default_value))
    except IOError:
        print("warning : config file not found, using default params.")
        pass
    except Exception:
        raise

    return params


def is_gandios():
    """Check if a VM is a GandiOS"""
    ret = False
    try:
        gandi_version = open('/etc/gandi/gandi_version').readline()
        gandi_version = gandi_version.strip()
        if gandi_version == 'gandios/stable':
            ret = True
    except IOError, e:
        # if file does not exist, do nothing, otherwise log
        if e.errno != 2:
            syslog.syslog('%s %s', e, str(e))
    except Exception, e:
        syslog.syslog('%s %s', e, str(e))
    return ret


def run_command(label, cmd):
    """Run a command, log its output, return its exit code"""
    syslog.syslog('%s (%s)' % (label, cmd,))
    command = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                stdout=subprocess.PIPE, stdin=None,
                shell=True)
    # wait until the process has completed
    (output, error) = command.communicate()
    ret = command.returncode

    # read the stdout/stderr lines
    if ret is not None:
        syslog.syslog('%s [%s]: %s stderr=%s' % (label, cmd, output, error))
    else:
        # log the return code for easy debugging
        syslog.syslog('%s [%s]: returned %s' % (label, cmd, output))
    return ret


def pre_add(device, fs_type, mountpoint):
    """Try to resize a disk to its block size before mount,
    eventually checking its filesystem"""

    resize_opts = ''

    # if it isn't a gandios, do nothing
    if not is_gandios():
        syslog.syslog('Not GandiOs, skipping pre-add')
        return

    if not fs_type == 'ext3' and not fs_type == 'ext4' \
       and not fs_type == 'ext2':
        syslog.syslog('Device %s does not contain an ext2/3/4 filesystem. \
            We simply mount it as is.' % (device,))
        return

    ## not yet resizing main disk in GandiOS. More test need to be done.
    #if device == '/dev/xvda1':
    #    try:
    #        run_command('auto-fsck', 'e2fsck -f -p %s' % (device,))
    #        run_command('auto-resize', 'resize2fs -f /%s' % (device, ))
    #    except:
    #        pass
    #    return

    # check if the disk is already mounted
    ret = run_command('check-mount', 'grep ^%s /proc/mounts' % (device,))

    if ret > 0:
        # device is not currently mounted
        #syslog.syslog("auto-fsck %s as not mounted" % (device, ))
        ret = run_command('auto-fsck', 'e2fsck -f -p %s' % (device,))
    else:
        # device is currently mounted, we try to umount
        try:
            #syslog.syslog("umount volume %s before fsck" % (device, ))
            run_command('umount volume', 'umount %s' % (device,))
            run_command('auto-fsck', 'e2fsck -f -p %s' % (device,))
        except:
            resize_opts = '-f'
            pass

    run_command('auto-resize', 'resize2fs %s %s' % (resize_opts, device))
    return


def on_add(device, mountpoint):
    """ mount a device on a defined mountpoint. """

    if re.match('/dev/xvda.?', device):
        return
    else:
        syslog.syslog('%s is ok to mount, going on' % device)

    mount_params = load_mount_params(config_file)

    if not os.path.exists(mountpoint):
        os.mkdir(mountpoint)

    if in_mtab(device):
        syslog.syslog('%s is already mounted.' % device)
        return

    if os.environ.get('ID_FS_TYPE'):
        cmd = 'mount -v -t %s -o %s %s %s' % (os.environ['ID_FS_TYPE'],
            mount_params, device, mountpoint)
    else:
        cmd = 'mount -v -o %s %s %s' % (mount_params, device, mountpoint)
    os.system(cmd)
    # if the mount command is not sucessfull, we remove the mount point
    if not os.path.ismount(mountpoint):
        cmd_result = os.rmdir(mountpoint)
        if cmd_result != 0:
            syslog.syslog('error when deleting ' + mountpoint)
    else:
        os.chown(mountpoint, 0, 1001)
        os.chmod(mountpoint, 0775)
        if not in_mtab(mountpoint):
            add_to_mtab(mountpoint)


def in_mtab(element):
    """ look in /etc/mtab to find if element is present. """
    mntfound = 0
    for line in open('/etc/mtab', 'r'):
        result = re.compile('%s' % element).search(line.rstrip(os.linesep))
        if result is not None:
            mntfound = 1
            break
    return mntfound


def add_to_mtab(mountpoint):
    """ add mountpoint to /etc/mtab when reading information
    from /proc/mounts.
    """
    for line in open('/proc/mounts', 'r'):
        result = re.compile('^%s' % mountpoint).search(line.rstrip(os.linesep))
        if result is not None:
            mtab_fd = open('/etc/mtab', 'a')
            mtab_fd.write(line.rstrip(os.linesep))
            mtab_fd.close()


def remove_mountpoint(mountpoint):
    """ remove a directory """

    if not os.path.ismount(mountpoint):
        cmd_result = os.rmdir(mountpoint)
        if cmd_result != 0:
            syslog.syslog('error when deleting ' + mountpoint)
    else:
        syslog.syslog('warn : something is mounted on ' + mountpoint)


def umount_device_pattern(device_pattern):
    """ umount all partitions according to device pattern. """

    patternc = re.compile(device_pattern)
    for line in open('/proc/mounts', 'r'):
        result = patternc.search(line.rstrip(os.linesep))
        if result is not None:
            cmd_result = os.system('umount -l ' + result.group(1))
            if cmd_result != 0:
                syslog.syslog('error when unmounting ' + result.group(1))


def on_delete(device, mountpoint):
    """ umount a device and remove a mountpoint. """

    # stopping services using the datadisk
    if os.path.exists(os.path.join(mountpoint, '.gandi')):
        if os.path.exists(os.path.join(mountpoint, '.gandi', 'services')):
            data = file(os.path.join(mountpoint, '.gandi', 'services'), 'r')
            services = data.read().rstrip(os.linesep)
            for service in services:
                # /etc/init.d/service stop
                command = subprocess.Popen('/etc/init.d/' + service + ' stop',
                                     stderr=subprocess.PIPE,
                                     stdout=subprocess.PIPE,
                                     stdin=None,
                                     shell=True)

                ret = command.returncode
                syslog.syslog('Trying to stop %s' % service)
                (output, error) = command.communicate()

                if ret is not None:
                    syslog.syslog('stopping %s, output : %s and error : %s' \
                        % (service, output, error))

    # first we unmount the specific device from the system
    umount_device_pattern('^(' + device + ') .*$')

    # we remove specific mountpoing
    if os.path.exists(mountpoint):
        remove_mountpoint(mountpoint)

    # ... and even mountpoint build the the label from udev
    if 'ID_FS_LABEL_SAFE' in os.environ:
        if len(os.environ['ID_FS_LABEL_SAFE']) > 0:
            mountpoint = datadisk_root + '/' + os.environ['ID_FS_LABEL_SAFE']
            if os.path.exists(mountpoint):
                remove_mountpoint(mountpoint)

    # then we remove existing partition on the device
    umount_device_pattern('^(' + device + '[0-9a-z]+)\s*.*')


def writable_root():
    if posix.access('/', posix.W_OK):
            return True
    return False


def main():
    syslog.openlog('gandi')

    if int(load_allow_mount(config_file)) == 0:
        sys.exit(0)

    if not writable_root():
        sys.exit(1)

    datadisk_root = load_disk_root(config_file)

    if not os.path.exists(datadisk_root):
        os.mkdir(datadisk_root)
        syslog.syslog('creating ' + datadisk_root)

    pattern = '.*/block/(.*)xvd.*'

    if 'ID_FS_TYPE' in os.environ and os.environ.get('ID_FS_TYPE') == 'swap':
        sys.exit(0)

    # we use the DEVPATH udev variable to get the device name
    if os.environ.get('DEVPATH') and re.match(pattern, os.environ['DEVPATH']):
        rawdevice = os.environ['DEVPATH']

        try:
            # we look in /sys tree to get minor and major information
            # for the device
            data = open('/sys' + rawdevice + '/dev', 'r')
            (major, minor) = data.read().rstrip(os.linesep).split(':')
        except IOError:
            sys.exit('Device ' + rawdevice + ' is not present.')

        # as long as the device is connected, we can find the real name with
        # minor and major info
        s = '\s+(' + major + ')\s+(' + minor + ')\s+.*(xvd[a-z0-9]+)$'
        sc = re.compile(s)
        # we build a return hash to know how many partitions we have for each
        # device
        partitions = {}
        device_label = ''
        for line in open('/proc/partitions', 'r'):
            result = sc.match(line)
            if result is not None:
                device_name = result.group(3)
                device = '/dev/' + device_name
            else:
                continue
            # device label as dict key (xvdb), list of partitions
            # ['xvdb', 'xvdb1'] etc. as value.
            device_label = re.match('^(xvd[a-z]+)', device_name).group(1)
            partitions[device_label] = partitions.get(device_label, [])
            partitions[device_label].append(device_name)

        # udev is providing us a label name if present
        if 'ID_FS_LABEL_SAFE' in os.environ and os.environ['ID_FS_LABEL_SAFE']:
            mountpoint = datadisk_root + '/' + os.environ['ID_FS_LABEL_SAFE']
        else:
            if 'ID_FS_LABEL_ENC' in os.environ and \
               os.environ['ID_FS_LABEL_ENC']:
                mountpoint = datadisk_root + '/' + \
                    os.environ['ID_FS_LABEL_ENC']
            else:
                # in some distribution, we have to ask e2label about it
                command = subprocess.Popen('/sbin/e2label ' + device,
                      shell=True,
                      stdin=subprocess.PIPE,
                      stdout=subprocess.PIPE,
                      stderr=subprocess.PIPE,
                      close_fds=True)
                ret = command.returncode
                (output, error) = command.communicate()
                if ret is None and not re.compile('^$').match(output):
                    mountpoint = datadisk_root + '/' + output
                else:
                    mountpoint = datadisk_root + '/' + \
                        os.path.basename(rawdevice)

        pattern = re.compile(mountpoint + ' ')
        for line in open('/proc/mounts', 'r'):
            match = pattern.search(line.rstrip(os.linesep))
            if match is not None:
                mountpoint = "%s_2" % mountpoint
                syslog.syslog('manage_data_disk: a disk with the same fs \
                   label is already present; mounting it in %s' % mountpoint)

        if re.match('add', os.environ['ACTION']) or \
           re.match('change', os.environ['ACTION']):
            fs_type = os.environ.get('ID_FS_TYPE')

            # we only resize disks without partitions
            if len(partitions.get(device_label, [])) == 1:
                pre_add(device, fs_type, mountpoint)
            on_add(device, mountpoint)

        if re.match('offline', os.environ['ACTION']):
            on_delete(device, mountpoint)

try:
    main()
except Exception, e:
    syslog.syslog('manage_data_disk: caught exception')
    import traceback
    syslog.syslog(traceback.format_exc())

# vim:ft=python:et:sw=4:ts=4:sta:tw=79:fileformat=unix
